import numpy as np

import torch

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.offsetbox import (AnchoredOffsetbox, HPacker, TextArea)
from matplotlib.collections import LineCollection
import matplotlib.colors as colors

from astropy.time import Time
import astropy.constants as consts
import astropy.units as u

from orbitize.kepler import calc_orbit


def plot_orbit(theta, data, timestep = None, text = False):
    """
    Plots the orbit based on the given theta values and data.

    Parameters:
    theta (numpy.ndarray): Array of parameter values.
    data (numpy.ndarray): Array of data points.
    """
    
    ra = data[0::2]
    dec = data[1::2]

    label_print = ['a', 'e', 'i', 'ω', 'Ω', 'τ', 'Mt']

    # Just plot the values different from 0
    non_zero_indices = np.where((ra != 0) & (dec != 0))
    ra = ra[non_zero_indices]
    dec = dec[non_zero_indices]

    if timestep is None:
        color = np.linspace(2002, 2019.9, 180)[non_zero_indices]
        print("Color is " + str(color))
    else:
        color = timestep

    fig, ax = plt.subplots() 

    shw = ax.scatter(ra, dec, c=color, cmap='winter', label="Simulated data")
    bar = plt.colorbar(shw)

    plt.xlabel('$\\Delta$ RA'); plt.ylabel('$\\Delta$ Dec')
    plt.axis('equal')
    plt.plot(0,0,marker="*",color='black',markersize=10, label = "Star", linestyle='')
    bar.set_label('Year')
    plt.legend()
    
    if text:
        box1 = TextArea('\n'.join([f"{label}: {theta_value:.3f}" for label, theta_value in zip(label_print[0::4], theta.tolist()[0::4])]))
        box2 = TextArea('\n'.join([f"{label}: {theta_value:.3f}" for label, theta_value in zip(label_print[1::4], theta.tolist()[1::4])]))
        box3 = TextArea('\n'.join([f"{label}: {theta_value:.3f}" for label, theta_value in zip(label_print[2::4], theta.tolist()[2::4])]))
        box4 = TextArea('\n'.join([f"{label}: {theta_value:.3f}\n" for label, theta_value in zip(label_print[3::4], theta.tolist()[3::4])]))
        box_title = TextArea("Parameters : \n")
        
        box = HPacker(children=[box_title, box1, box2, box3, box4],
                    align="left",
                    pad=0, sep=5) 
        
        anchored_box = AnchoredOffsetbox(loc='lower left',
                                        child=box, pad=0.5,
                                        frameon=True,
                                        bbox_to_anchor=(0., 1.02),
                                        bbox_transform=ax.transAxes,
                                        borderpad=0.,)
        ax.add_artist(anchored_box)
        fig.subplots_adjust(top=0.8)
    plt.show()


def ra_dec_plot(estimator, prior, x, num_samples=2000):
    with torch.no_grad():
        samples = estimator.flow(x.cuda()).sample((num_samples,)).cpu()
        mask = torch.all((samples >= -1) & (samples <= 1), dim=1, keepdim=True)
        samples = samples[mask.squeeze()]
        samples = prior.post_process(samples)

    sma, ecc, inc, aop, pan, tau, mtot = samples.T.numpy()

    inc = np.radians(inc)
    aop = np.radians(aop)
    pan = np.radians(pan)

    t_obs = np.linspace(2002, 2019.9, 180) # should add the discretization as an argument !
    t = np.linspace(2000, 2029.9, 180) # should add the discretization as an argument !
    observation_epochs = Time(t, format='decimalyear').mjd

    ra, dec, _ = calc_orbit(observation_epochs, 
                            sma, 
                            ecc, 
                            inc, 
                            aop, 
                            pan, 
                            tau, 
                            100, 
                            mtot, 
                            use_c = False, # Should be set to True on the cluster
                            tau_ref_epoch=50000)
    ra = ra.T
    dec = dec.T

    obs = x*2e4 # should add the scaling factor as an argument ! 
    ra_obs = obs[0::2]
    dec_obs = obs[1::2]

    non_zero_indices = np.where((ra_obs != 0) & (dec_obs != 0))
    ra_obs = ra_obs[non_zero_indices]
    dec_obs = dec_obs[non_zero_indices]
    t_obs= t_obs[non_zero_indices]

    fig, axs = plt.subplots(2)

    axs[0].fill_between(t,
                    np.quantile(ra, 0.005, axis=0),
                    np.quantile(ra, 0.995, axis=0),
                    color = "#1f77b4",
                    alpha=0.3,
                    label = "99\%")
    axs[0].fill_between(t,
                    np.quantile(ra, 0.025, axis=0),
                    np.quantile(ra, 0.975, axis=0),
                    color = "#1f77b4",
                    alpha=0.4,
                    label = "95\%")
    axs[0].fill_between(t,
                    np.quantile(ra, 0.1, axis=0),
                    np.quantile(ra, 0.9, axis=0),
                    color = "#1f77b4",
                    alpha=0.5,
                    label = "80\%") 

    axs[1].fill_between(t,
                    np.quantile(dec, 0.005, axis=0),
                    np.quantile(dec, 0.995, axis=0),
                    color = "#1f77b4",
                    alpha=0.3,
                    label = "99\%")
    axs[1].fill_between(t,
                    np.quantile(dec, 0.025, axis=0),
                    np.quantile(dec, 0.975, axis=0),
                    color = "#1f77b4",
                    alpha=0.4,
                    label = "95\%")
    axs[1].fill_between(t,
                    np.quantile(dec, 0.1, axis=0),
                    np.quantile(dec, 0.9, axis=0),
                    color = "#1f77b4",
                    alpha=0.5,
                    label = "80\%")

    axs[0].scatter(t_obs, ra_obs, marker="x", color = "r", zorder=1, label = "Observations")
    axs[0].set_ylabel('$\\Delta$ RA')
    axs[0].set_xlim(2000, 2032)
    axs[0].grid()
    axs[0].legend(bbox_to_anchor=(1.05, 1),
                            loc='upper left', borderaxespad=0., title = "Confidence interval")

    axs[1].scatter(t_obs, dec_obs, marker="x", color = "r", zorder=1, label = "Observations")
    axs[1].set_ylabel('$\\Delta$ DEC')
    axs[1].set_xlim(2000, 2032)
    axs[1].grid()

    axs[1].set_xlabel('Time')

    plt.show()


# modified from https://orbitize.readthedocs.io/en/latest/_modules/orbitize/plot.html#plot_orbits
def plot_orbits(estimator, prior, x, num_orbits=100):

    fig, ax = plt.subplots()
    plt.xlabel('$\\Delta$ RA'); plt.ylabel('$\\Delta$ Dec')
    plt.axis('equal')
    plt.plot(0,0,marker="*",color='black',markersize=10)

    with torch.no_grad():
        samples = estimator.flow(x.cuda()).sample((int(num_orbits * 1.5),)).cpu() # Increase the number of samples to account for the ones that will be removed
        mask = torch.all((samples >= -1) & (samples <= 1), dim=1, keepdim=True) # Remove samples with values outside [-1, 1]
        samples = samples[mask.squeeze()]
        samples = prior.post_process(samples)
        samples = samples[:num_orbits] # Keep only the first num_orbits samples
    
    raoff = np.zeros((100, 100))
    deoff = np.zeros((100, 100))
    epochs = np.zeros((100, 100))

    for i in range(0,100):
        sma, ecc, inc, aop, pan, tau, mtot = samples[i].numpy()

        inc = np.radians(inc)
        aop = np.radians(aop)
        pan = np.radians(pan)
        
        period = np.sqrt(4*np.pi**2.0*(sma*u.AU)**3/(consts.G*(mtot*u.Msun)))
        period = period.to(u.day).value
        
        epochs[i,:] = np.linspace(50000, float(50000 + period), 100)
                                                    
        ra, dec, _ = calc_orbit(epochs[i,:], 
                                sma, 
                                ecc, 
                                inc, 
                                aop, 
                                pan, 
                                tau, 
                                100, 
                                mtot,
                                use_c=False, # Change to true on the cluster
                                tau_ref_epoch=50000)
        
        raoff[i, :] = ra
        deoff[i, :] = dec

    min_cbar_date = np.min(epochs)
    max_cbar_date = np.max(epochs)

    if max_cbar_date - min_cbar_date > 1000 * 365.25:
        max_cbar_date = min_cbar_date + 1000 * 365.25

    norm = mpl.colors.Normalize(vmin=min_cbar_date, vmax=max_cbar_date)
    norm_yr = mpl.colors.Normalize(
        vmin=Time(min_cbar_date, format='mjd').decimalyear,
        vmax=Time(max_cbar_date, format='mjd').decimalyear
    )

    cmap = mpl.cm.hsv
    cmap = colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=0.0, b=0.7),
        cmap(np.linspace(0, 0.85, 10000))
    )

    for j in range(0,100):
        points = np.array([raoff[j, :], deoff[j, :]]).T.reshape(-1, 1, 2)
        segments = np.concatenate([points[:-1], points[1:]], axis=1)
        lc = LineCollection(
            segments, cmap=cmap, norm=norm, linewidth=1.0
        )
        lc.set_array(epochs[j, :])
        ax.add_collection(lc)  

    obs = x*2e4 # should add the scaling factor as an argument ! 
    ra_obs = obs[0::2]
    dec_obs = obs[1::2]

    plt.scatter(ra_obs, dec_obs, color = "black",  marker="x",zorder=10, label = "Observations")
    plt.plot(0,0,marker="*", linestyle='',color='black',markersize=10, zorder=20, label = "Star")
    plt.legend()
    cbar_ax = fig.add_axes([0.92, 0.15, 0.015, 0.7])
    cbar = mpl.colorbar.ColorbarBase(cbar_ax, cmap=cmap, norm=norm_yr, orientation='vertical', label='Epoch [year]')

    plt.show()    


def corner_plot(estimator, prior, LOWER, UPPER):
    with torch.no_grad():
        samples = estimator.flow(testset[i][1].cuda()).sample((2**16,)).cpu()
        mask = torch.all((samples >= -1) & (samples <= 1), dim=1, keepdim=True) # Remove samples with values outside [-1, 1]
        samples = samples[mask.squeeze()]
        samples = prior.post_process(samples)
        
    LOWER = torch.tensor([8.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
    UPPER = torch.tensor([200.0, 0.9, 180.0, 360.0, 360.0, 1.0, 3.0])

    LABELS = [r'$a$', r'$e$', r'$i$',
            r'$\omega$', r'$\Omega$',
            r'$\tau$', r'$M_T$']

    plt.rcParams.update(nice_rc(latex=True))  # nicer plot settings

    fig = corner(
        samples,
        domain=(LOWER, UPPER),
        #smooth=2,
        labels=LABELS,
        legend=r'$p_\phi(\theta | x^*)$',
    )

    mark_point(fig, testset[i][0])
    plt.show()
