import torch
import torch.nn as nn
from torch import Tensor

from lampe.nn import ResMLP
from lampe.inference import NPE

class NPEWithResMLP(nn.Module): 
    def __init__(self, 
                 num_obs, 
                 embedding_output_dim,
                 embedding_hidden_features,
                 activation,
                 transforms,
                 flow,
                 NPE_hidden_features,
                 mass_as_input = False):
        
        super().__init__()
        
        if mass_as_input: 
            num_obs = num_obs * 2 + 1
            parameters = 6
        else:
            num_obs = num_obs * 2
            parameters = 7
                   
        self.embedding = nn.Sequential(ResMLP(num_obs,
                                              embedding_output_dim,
                                              hidden_features = [512,512,256,256,256,128,128,128,128,128],
                                              activation = activation))
        
        
        self.npe = NPE(parameters,
                       embedding_output_dim, 
                       transforms = transforms, 
                       build = flow, 
                       hidden_features = NPE_hidden_features, 
                       activation = activation)
        
        self.mass_as_input = mass_as_input
        
    def forward(self, theta: Tensor, x: Tensor) -> Tensor:
        if self.mass_as_input:
            theta = theta[:,:-2].contiguous() # remove plx & mtot
        else:
            theta = theta[:, [0, 1, 2, 3, 4, 5, 7]].contiguous() # remove plx
        return self.npe(theta, self.embedding(x))

    def flow(self, x: Tensor): 
        # The tanh function is used to ensure that the output stays between -1 and 1
        return self.npe.flow(self.embedding(x))
    
class NPEWithDeepSet(nn.Module):
    def __init__(self, 
                 num_obs, 
                 embedding_output_dim,
                 embedding_hidden_features,
                 activation,
                 transforms,
                 flow,
                 NPE_hidden_features,
                 mass_as_input = False):
        
        super().__init__()
        
        if mass_as_input: 
            input = 19
            parameters = 6
        else:
            input = 18
            parameters = 7
              
                
        #self.feature_extractor = nn.Sequential(
        #    nn.Linear(input, embedding_hidden_features),
        #    activation(),
        #    nn.Linear(embedding_hidden_features, embedding_hidden_features),
        #    activation(),
        #    nn.Linear(embedding_hidden_features, embedding_hidden_features)
        #)

        #self.regressor = nn.Sequential(
        #    nn.Linear(embedding_hidden_features, embedding_hidden_features),
        #    activation(),
        #    nn.Linear(embedding_hidden_features, embedding_hidden_features),
        #    activation(),
        #    nn.Linear(embedding_hidden_features, embedding_hidden_features),
        #    activation(),
        #    nn.Linear(embedding_hidden_features, embedding_output_dim),
        #) 

        self.feature_extractor = nn.Sequential(
#             nn.Linear(3, 16),
            nn.Linear(input, 16),
            nn.ELU(inplace=True),
            nn.Linear(16, 32),
            nn.ELU(inplace=True),
            nn.Linear(32, 64)
        )

        self.regressor = nn.Sequential(
            nn.Linear(64, 128),
            nn.ELU(inplace=True),
            nn.Linear(128, 256),
            nn.ELU(inplace=True),
            nn.Linear(256, 128),
            nn.ELU(inplace=True),
            nn.Linear(128, 64),
        )  
        
        self.npe = NPE(parameters, 
                       embedding_output_dim, 
                       transforms = transforms, 
                       build = flow, 
                       hidden_features = NPE_hidden_features, 
                       activation = activation)

        self.mass_as_input = mass_as_input
        
    def forward(self, theta: Tensor, inputs: Tensor) -> Tensor:
        
        if self.mass_as_input:
            theta = theta[:,:-2]
            
        inputs = inputs.to(torch.float32)

        # If the first element of the input is 0, it means it should be masked
        mask = (inputs[:,:,0] != 0)


        x = self.feature_extractor(inputs)
        x *= mask.unsqueeze(-1)

        x = x.sum(dim=1)
        x = self.regressor(x)

        return self.npe(theta, x)
    
    def flow(self, inputs: Tensor):
        
        inputs = inputs.to(torch.float32)

        # If the first element of the input is 0, it means it should be masked
        mask = (inputs[:,:,0] != 0)

        x = self.feature_extractor(inputs)
        x *= mask.unsqueeze(-1)


        x = x.sum(dim=1)
        x = self.regressor(x)
        
        # The tanh function is used to ensure that the output stays between -1 and 1
        return self.npe.flow(x)